####
#
# Dreamy Utilities
# Copyright (C) (2020) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from dreamy_utilities.Text import Stringify

# Standard packages.

from requests import get, Session
from typing import Optional
from urllib.parse import urlparse

# Non-standard packages.

from bs4 import BeautifulSoup
import tldextract

#
#
#
# Functions.
#
#
#

def DownloadPage(URL: str, session: Optional[Session] = None) -> Optional[str]:

    ##
    #
    # Downloads a webpage and returns its code.
    #
    # @param URL     The URL.
    # @param session The Session object to be used for the download.
    #
    # @return The code of the downloaded page. **None**, if an error has occurred.
    #
    ##

    if not URL:
        return None

    response = session.get(URL) if session else get(URL)
    if not response:
        return None

    return Stringify(response.content)

def DownloadSoup(
    URL: str,
    session: Optional[Session] = None,
    parser: str = "html.parser"
) -> Optional[BeautifulSoup]:

    ##
    #
    # Downloads a webpage and returns it as a soup of tags.
    #
    # @param URL     The URL.
    # @param session The Session object to be used for the download.
    # @param parser  The HTML parser to be used.
    #
    # @return The tag soup. **None**, if an error has occurred.
    #
    ##

    code = DownloadPage(URL, session)
    if not code:
        return None

    return BeautifulSoup(code, parser)

def GetHostname(URL: str) -> str:

    ##
    #
    # Retrieves hostname from a URL ("protocol://a.b.com/1/2/3/" returns "b.com").
    #
    # @param URL The URL.
    #
    # @return The hostname extracted from the input URL.
    #
    ##


    if not URL:
        return URL

    parts = tldextract.extract(URL)
    return f"{parts.domain}.{parts.suffix}"

def GetSiteURL(URL: str) -> str:

    ##
    #
    # Returns the URL to the main site ("protocol://a.b.com/1/2/3/" returns "protocol://a.b.com").
    #
    # @param URL The URL.
    #
    # @return The URL to the main site.
    #
    ##

    if not URL:
        return URL

    URL = urlparse(URL)

    return f"{URL.scheme}://{URL.netloc}"


####
#
# Dreamy Utilities
# Copyright (C) (2020) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Standard packages.

from typing import Tuple

#
#
#
# Functions.
#
#
#

def GetDimensionsToFit(
    dimensions: Tuple[int, int],
    containerDimensions: Tuple[int, int]
) -> Tuple[int, int]:

    ##
    #
    # Proportionally scales a rectangle to fit inside another rectangle.
    #
    # @param dimensions          The dimensions of the rectangle.
    # @param containerDimensions The dimensions of the outside rectangle.
    #
    # @return A tuple containing new dimensions for the inside rectangle: (W, H).
    #
    ##

    width, height = dimensions

    if (not width) or (not height):
        return (0, 0)

    containerWidth, containerHeight = containerDimensions

    scale = min(
        containerWidth / width,
        containerHeight / height
    )

    return (
        int(scale * width),
        int(scale * height)
    )


####
#
# Dreamy Utilities
# Copyright (C) (2020) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Standard packages.

import os
from typing import List, Optional, Tuple

# Non-standard packages.

import colorama
import termtables

#
#
#
# Classes.
#
#
#

##
#
# Represents textual user interface.
#
##

class Interface:

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        colorama.init()

    def GrabUserAttention(self) -> None:

        ##
        #
        # Attempts to grab the user's attention by making a "beep" sound. (On Windows it also makes
        # the taskbar button flash.)
        #
        ##

        print("\a")

        if "nt" == os.name:

            import ctypes

            ctypes.windll.user32.FlashWindow(
                ctypes.windll.kernel32.GetConsoleWindow(),
                True
            )

    def Text(
        self,
        text: str,
        section: bool = False,
        clearLine: bool = False,
        end: str = "\n",
        color: Optional[str] = None,
        bold: bool = False
    ) -> None:

        ##
        #
        # Prints a line of text.
        #
        # @param text      The text to be printed.
        # @param section   Should we print an empty line before this text?
        # @param clearLine Should we clear the current line before printing?
        # @param end       The string printed immediately after the given text.
        # @param color     The color of the text.
        # @param bold      Should the printed text be bold?
        #
        ##

        if clearLine:
            self.ClearLine()

        if section:
            self.EmptyLine()

        print(
            self._GetFormattedText(text, color, bold),
            end = end
        )

    def Comment(
        self,
        text: str,
        section: bool = False,
        clearLine: bool = False,
        end: str = "\n"
    ) -> None:

        ##
        #
        # Prints a comment.
        #
        # @param text      The text to be printed.
        # @param section   Should we print an empty line before this text?
        # @param clearLine Should we clear the current line before printing?
        # @param end       The string printed immediately after the given text.
        #
        ##

        self.Text(f"# {text}", section, clearLine, end)

    def Process(
        self,
        text: str,
        section: bool = False,
        clearLine: bool = False,
        end: str = "\n"
    ) -> None:

        ##
        #
        # Prints a description of an ongoing process.
        #
        # @param text      The text to be printed.
        # @param section   Should we print an empty line before this text?
        # @param clearLine Should we clear the current line before printing?
        # @param end       The string printed immediately after the given text.
        #
        ##

        self.Text(f"> {text}", section, clearLine, end, color = "cyan", bold = True)

    def Error(
        self,
        text: str,
        section: bool = False,
        clearLine: bool = False,
        end: str = "\n"
    ) -> None:

        ##
        #
        # Prints an error of an ongoing process.
        #
        # @param text      The text to be printed.
        # @param section   Should we print an empty line before this text?
        # @param clearLine Should we clear the current line before printing?
        # @param end       The string printed immediately after the given text.
        #
        ##

        self.Text(f"! {text}", section, clearLine, end, color = "red")

    def Notice(
        self,
        text: str,
        section: bool = False,
        clearLine: bool = False,
        end: str = "\n"
    ) -> None:

        ##
        #
        # Prints some optional information.
        #
        # @param text      The text to be printed.
        # @param section   Should we print an empty line before this text?
        # @param clearLine Should we clear the current line before printing?
        # @param end       The string printed immediately after the given text.
        #
        ##

        self.Text(f"! Notice: {text}", section, clearLine, end, color = "green")

    def EmptyLine(self) -> None:

        ##
        #
        # Prints an empty line.
        #
        ##

        print()

    def ClearLine(self) -> None:

        ##
        #
        # Clears the current line.
        #
        ##

        print("\r", end = "")

    def LineBreak(self) -> None:

        ##
        #
        # Prints a line break.
        #
        ##

        self.Text(
            79 * "-",
            section = True,
            color = "cyan"
        )

    def ProgressBar(
        self,
        progress: int,
        total: int,
        length: int,
        description: str,
        clearLine: bool = True
    ) -> None:

        ##
        #
        # Prints a progress bar.
        #
        # @param progress    Current progress.
        # @param total       The maximum value of progress.
        # @param length      The length of the progress bar.
        # @param description The description.
        # @param clearLine   Should we clear the line before printing?
        #
        ##

        if clearLine:
            self.ClearLine()

        proportion = float(progress) / float(total)

        filledPart = int(proportion * length)
        remainingPart = length - filledPart

        bar = (filledPart * "█") + (remainingPart * "-")

        print(f"{description}: |{bar}|", end = "")

    def Table(
        self,
        data: List[List[str]],
        padding: Tuple[int, int] = (0, 1),
        alignment: str = "rl"
    ):

        termtables.print(
            data,
            style = termtables.styles.thin,
            padding = padding,
            alignment = alignment
        )

    @staticmethod
    def _GetColorMarker(
        color: str
    ) -> str:

        ##
        #
        # Returns a color marker for given color.
        #
        # @param color The name of the color.
        #
        # @return The color marker.
        #
        ##

        if color not in Interface._COLORS:
            return Interface._GetColorMarker("reset")

        return f"\u001b[{str(Interface._COLORS[color])}m"

    @staticmethod
    def _GetFormattedText(
        text: str,
        color: Optional[str] = None,
        bold: bool = False
    ) -> str:

        ##
        #
        # Formats text in a given way.
        #
        # @param text  The text to be formatted.
        # @param color The color of the text.
        # @param bold  The weight of the text.
        #
        # @return A formatted string, ready for printing.
        #
        ##

        if color:
            text = Interface._GetTextMarker(color) + text

        if bold:
            text = Interface._GetTextMarker("bold") + text

        text += Interface._GetTextMarker("reset")

        return text

    @staticmethod
    def _GetTextMarker(
        marker: str
    ) -> str:

        ##
        #
        # Returns an ANSI text formatting marker.
        #
        # @param marker The name of the marker.
        #
        # @return The marker.
        #
        ##

        if marker not in Interface._MARKERS:
            return Interface._GetTextMarker("reset")

        return f"\u001b[{str(Interface._MARKERS[marker])}m"

    _MARKERS = {
        "reset"  : 0 ,
        "bold"   : 1 ,
        "black"  : 30,
        "red"    : 31,
        "green"  : 32,
        "yellow" : 33,
        "blue"   : 34,
        "magenta": 35,
        "cyan"   : 36,
        "white"  : 37,
    }


####
#
# Dreamy Utilities
# Copyright (C) (2020) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Standard packages.

import os
from os.path import expandvars, isfile
from pathlib import Path
import shutil
from string import ascii_letters, digits
import sys
from typing import List, Optional, Union
from uuid import uuid4

#
#
#
# Constants.
#
#
#

VALID_FILE_NAME_CHARACTERS = f"-',_.()[] {ascii_letters}{digits}"

#
#
#
# Functions.
#
#
#

def AddToPATH(path: Union[str, Path]) -> None:

    ##
    #
    # Adds a directory path to the PATH variable.
    #
    # @param path The directory path to be added.
    #
    ##

    if not path.is_dir():
        return

    sys.path.insert(0, str(path))

def CopyTree(sourceDirectoryPath: Path, destinationDirectoryPath: Path) -> None:

    ##
    #
    # Copies a directory, along with its contents.
    #
    # @param sourceDirectoryPath      The directory to be copied.
    # @param destinationDirectoryPath The directory in which the copy will be placed.
    #
    ##

    if not sourceDirectoryPath.is_dir():
        return

    destinationDirectoryPath.parent.mkdir(parents = True, exist_ok = True)

    shutil.copytree(sourceDirectoryPath, destinationDirectoryPath)

def FindExecutable(
    fileName: str,
    applicationName: Optional[str] = None,
    relativePath: Optional[str] = None
) -> Optional[Path]:

    ##
    #
    # Returns the absolute path to an executable.
    #
    # @param fileName        The name of the executable (without file name extension).
    #                        For example: "soffice".
    # @param applicationName The name of the application the executable belongs to.
    #                        For example: "LibreOffice".
    # @param relativePath    The relative path to the executable, inside the application's
    #                        directory. For example: "program".
    #
    # @return The absolute file path, or **None**.
    #
    ##

    path = shutil.which(fileName)
    if path:
        return Path(path)

    path = f"/usr/bin/{fileName}"
    if isfile(path):
        return Path(path)

    path = f"/bin/{fileName}"
    if isfile(path):
        return Path(path)

    localExecutablePath = f"{fileName}.exe"
    if relativePath:
        localExecutablePath = f"{relativePath}\\" + localExecutablePath
    if applicationName:
        localExecutablePath = f"{applicationName}\\" + localExecutablePath

    path = expandvars(f"%ProgramW6432%\\{localExecutablePath}")
    if isfile(path):
        return Path(path)

    path = expandvars(f"%ProgramFiles(x86)%\\{localExecutablePath}")
    if isfile(path):
        return Path(path)

    return None

def FindFiles(
    directoryPath: Optional[Union[str, Path]] = None,
    recursive: bool = True,
    suffixes: Optional[List[str]] = None
) -> List[Path]:

    ##
    #
    # Finds all the files in given directory.
    #
    # @param directoryPath The path to the directory. Optional.
    # @param recursive     Should we search the sub-directories as well?
    # @param suffixes      A list of accepted file name suffixes. Optional.
    #
    # @return A list of file paths.
    #
    ##

    # Process the directory path; set it to CWD if no path is provided.

    directoryPath = Path(directoryPath) if directoryPath else Path()

    # Find all files in the directory.

    searchExpression = "**/*" if recursive else "*"
    filePaths = [x for x in directoryPath.glob(searchExpression) if x.is_file()]

    # Process suffixes (make them all lowercase), then use them to filter the list of file paths.

    if suffixes:

        suffixes = [x.lower() for x in suffixes]
        filePaths = [x for x in filePaths if x.suffix.lower() in suffixes]

    # Return.

    return filePaths

def GetSanitizedFileName(string: str) -> str:

    ##
    #
    # Generates a valid file name from any string.
    #
    # @param string The input string.
    #
    # @return Sanitized input string.
    #
    ##

    return "".join(x for x in string if x in VALID_FILE_NAME_CHARACTERS)

def GetUniqueFileName() -> str:

    ##
    #
    # Generates a unique file name. Useful for creating temporary files.
    #
    # @return A unique file name.
    #
    ##

    return uuid4().hex

def ReadTextFile(
    filePath: Union[str, Path],
    lines: bool = False
) -> Union[Optional[str], Optional[List[str]]]:

    ##
    #
    # Reads the contents of a file to a string.
    #
    # @param filePath The file path.
    # @param lines    Should we return a list of lines?
    #
    # @return File contents - as either one long string, or a list of strings - or **None**.
    #
    ##

    # Process the file path.

    filePath = Path(filePath)

    if (not filePath) or not filePath.is_file():
        return None

    # Open and read the file.

    try:

        with open(filePath, "r", encoding = "utf-8") as file:
            return file.read() if (not lines) else file.readlines()

    except OSError:

        return None

def RemoveEmptyDirectories(directoryPath: Path) -> None:

    ##
    #
    # Removes all empty subdirectories within given directory.
    #
    # @param directoryPath The directory.
    #
    ##

    for path in [x for x in directoryPath.glob("*") if x.is_dir()]:

        if not len(os.listdir(path)):
            shutil.rmtree(path)

def WriteTextFile(filePath: Union[str, Path], content: str) -> bool:

    ##
    #
    # Writes a string to file, creating it or overwriting if necessary. The directory tree in which
    # the file is meant to reside will be created, if necessary.
    #
    # @param filePath The file path.
    # @param content  The file content.
    #
    # @return **True** if the file was written successfully, **False** otherwise.
    #
    ##

    # Process the file path.

    filePath = Path(filePath)

    if not filePath:
        return False

    # Create the directory tree.

    filePath.parent.mkdir(parents = True, exist_ok = True)

    # Write the file.

    try:

        with open(filePath, "w", encoding = "utf-8") as file:
            file.write(content)

        return True

    except OSError:

        return False


####
#
# Dreamy Utilities
# Copyright (C) (2020) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Globals.
#
#
#

ApplicationName = "Dreamy Utilities"
ApplicationVersion = "1.0.0"
ApplicationShortDescription = "A collection of various utilities. "
ApplicationURL = "https://github.com/DreamCobbler/dreamy-utilities"

CreatorName = "Benedykt Synakiewicz"
CreatorContact = "dreamcobbler@outlook.com"


####
#
# Dreamy Utilities
# Copyright (C) (2020) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Standard packages.

from datetime import date, datetime
import re
from typing import Any, List, Optional, Tuple

# Non-standard packages.

from babel.dates import format_date
from babel.numbers import format_decimal
from titlecase import titlecase

#
#
#
# Constants.
#
#
#

VALID_ROMAN_NUMERAL_CHARACTERS = [
    "I",
    "V",
    "X",
    "L",
    "C",
    "D",
    "M",
]

#
#
#
# Functions.
#
#
#

def Bytify(value: Any) -> bytes:

    ##
    #
    # Converts any value to bytes.
    #
    # @param value The data to be bytified.
    #
    # @return Bytified input value.
    #
    ##

    if value is None:
        return b""

    elif isinstance(value, bytes):
        return value

    else:
        return bytes(value, encoding = "utf-8")

def DeprettifyAmount(amount: str) -> Tuple[int, int]:

    ##
    #
    # Converts an "amount" (i.e. something like "1,036/2,316" to a tuple of integers; in this case
    # it would be (1036, 2316)).
    #
    # @param amount The textual amount.
    #
    # @return A tuple of integers.
    #
    ##

    DEFAULT_RETURN_VALUE = (0, 0)

    if not amount:
        return DEFAULT_RETURN_VALUE

    amount = amount.strip().split("/")
    if len(amount) < 2:
        return DEFAULT_RETURN_VALUE

    return (
        DeprettifyNumber(amount[0]),
        DeprettifyNumber(amount[1])
    )

def DeprettifyNumber(number: str) -> int:

    ##
    #
    # Converts a "pretty" number to integer. (I.e.: "1,036" to 1036).
    #
    # @param number The input number.
    #
    # @return Integer version of the input number.
    #
    ##

    DEFAULT_RETURN_VALUE = 0

    if not number:
        return DEFAULT_RETURN_VALUE

    number = number.strip().replace(",", "")

    try:

        number = int(number)

    except ValueError:

        return DEFAULT_RETURN_VALUE

    return number

def FillTemplate(values: Any, template: str) -> str:

    ##
    #
    # Fills in a template string using values coming from an object, or a namespace, or anything
    # else iterable. Variable names in the template string need to be preceded and followed by three
    # "at" signs; example: @@@VariableName@@@.
    #
    # @param values   Iterable values.
    # @param template Template code.
    #
    # @return Template code with variables filled in.
    #
    ##

    valuesDictionary = vars(values) if not isinstance(values, dict) else values
    availableValues = [x for x in valuesDictionary if not x.startswith("_")]

    for value in availableValues:
        template = template.replace(f"@@@{value}@@@", str(valuesDictionary[value]))

    return template

def GetCurrentDate() -> str:

    ##
    #
    # Returns today's date, in ISO 8601 format (YYYY-MM-DD).
    #
    # @return Today's date.
    #
    ##

    return date.today().isoformat()

def GetDateFromTimestamp(timestamp: str) -> str:

    ##
    #
    # Converts Unix timestamp to ISO 8601 date.
    #
    # @param timestamp The input timestamp.
    #
    # @return The input date converted to ISO 8601 date format (YYYY-MM-DD).
    #
    ##

    return date.fromtimestamp(timestamp).isoformat()

def GetLevenshteinDistance(firstString: str, secondString: str) -> int:

    ##
    #
    # Calculates the Levenshtein distance between a pair of strings.
    #
    # @param firstString  The first string.
    # @param secondString The second string.
    #
    # @return The distance between the two strings.
    #
    ##

    if (not firstString) and (not secondString):
        return 0

    sizeH = len(firstString) + 1
    sizeV = len(secondString) + 1

    matrix = [[0 for i in range(sizeV)] for j in range(sizeH)]

    for x in range(sizeH):
        matrix[x][0] = x
    for y in range(sizeV):
        matrix[0][y] = y

    for x in range(1, sizeH):

        for y in range(1, sizeV):

            if firstString[x - 1] == secondString[y - 1]:

                matrix[x][y] = min(
                    matrix[x - 1][y    ] + 1,
                    matrix[x - 1][y - 1]    ,
                    matrix[x    ][y - 1] + 1
                )

            else:

                matrix[x][y] = min(
                    matrix[x - 1][y    ] + 1,
                    matrix[x - 1][y - 1] + 1,
                    matrix[x    ][y - 1] + 1
                )

    return matrix[sizeH - 1][sizeV - 1]

def GetLongestLeadingSubstring(strings: List[str]) -> Optional[str]:

    ##
    #
    # Retrieves the longest leading (i.e. starting) substring from a list of strings.
    #
    # @param strings A list of strings.
    #
    # @return The longest leading substring, or **None**.
    #
    ##

    if not strings:
        return None

    shortestStringLength = 0
    for string in strings:
        shortestStringLength = max(shortestStringLength, len(string))

    if not shortestStringLength:
        return None

    longestLeadingSubstring = ""
    currentCharacter = strings[0][0]

    for characterIndex in range(0, shortestStringLength):

        allStringsMatchCharacter = True

        for string in strings:

            if string[characterIndex] != currentCharacter:
                allStringsMatchCharacter = False
                break

        if not allStringsMatchCharacter:
            break

        longestLeadingSubstring += currentCharacter

        if characterIndex != shortestStringLength - 2:
            currentCharacter = strings[0][characterIndex + 1]

    return longestLeadingSubstring

def IsRomanNumeral(text: str) -> bool:

    ##
    #
    # Checks if given string is a Roman numeral.
    #
    # @param text The input string.
    #
    # @return **True** if the string is a Roman numeral, **False** otherwise.
    #
    ##

    if (not text) or any(x not in VALID_ROMAN_NUMERAL_CHARACTERS for x in text.strip()):
        return False

    return True

def IsStringEmpty(text: str) -> bool:

    ##
    #
    # Checks whether a string is empty. A string is considered empty if its length is 0, or when
    # it's composed solely of whitespace, or when it doesn't exist at all.
    #
    # @param text The input string.
    #
    # @return **True** if the string is empty, **False** otherwise.
    #
    ##

    if (not text) or text.isspace():
        return True

    return False

def PrettifyDate(
    date: str,
    inputFormat: str = "%Y-%m-%d",
    locale: str = "en"
) -> str:

    ##
    #
    # Returns a nicely formatted date.
    #
    # @param date        The input date.
    # @param inputFormat The format of the input date.
    # @param locale      The locale to be used for formatting.
    #
    # @return Prettified input date.
    #
    ##

    if "?" == date:
        return date

    return format_date(
        datetime.strptime(date, inputFormat),
        locale = locale
    )

def PrettifyNumber(
    number: int,
    locale: str = "en",
    isZeroSpecial: bool = False
) -> str:

    ##
    #
    # Returns a nicely formatted number.
    #
    # @param number        The input number.
    # @param locale        The locale to be used for formatting.
    # @param isZeroSpecial Should we replace zero with a question mark (0 -> "?")?
    #
    # @return Prettified input number.
    #
    ##

    if (0 == number) and isZeroSpecial:
        return "?"

    return format_decimal(
        number,
        locale = locale
    )

def PrettifyTitle(title: str, removeContext: bool) -> str:

    ##
    #
    # Returns a prettified title.
    #
    # @param title         The input title.
    # @param removeContext Should we remove the context? (Like "Chapter 3: ".)
    #
    # @return Prettified input title.
    #
    ##

    if not title:
        return title

    if removeContext:

        POSSIBLE_PREFIXES = [
            "Chapter \d+",
            "Ch\. \d+",
            "Update \d+",
            "Part \d+",
        ]

        POSSIBLE_POSTFIXES = [
            "Finale",
            "Final Part",
            "Final Update",
            "Final",
            "Chapter \d+",
            "Ch\. \d+",
            "Update \d+",
            "Part \d+",
        ]

        POSSIBLE_PREFIXES_JOINED = "|".join(POSSIBLE_PREFIXES)
        POSSIBLE_POSTFIXES_JOINED = "|".join(POSSIBLE_POSTFIXES)

        title = re.sub(
            f"^\[?\(?({POSSIBLE_PREFIXES_JOINED}|\d+)\)?\]?:?\.?",
            "",
            title,
            flags = re.IGNORECASE | re.MULTILINE
        )

        title = re.sub(
            f"\(?\[?({POSSIBLE_POSTFIXES_JOINED})\)?\]?$",
            "",
            title,
            flags = re.IGNORECASE | re.MULTILINE
        )

        pass

    title = titlecase(title)
    title = title.strip()

    if IsStringEmpty(title):
        title = ""

    return title

def SeparateSubtitle(title: str) -> Optional[str]:

    ##
    #
    # Retrieves the proper subtitle of the story ("aaa: bbbbb" will return "bbbbb").
    #
    # @param title The title as it was retrieved.
    #
    # @return The subtitle.
    #
    ##

    if not title:
        return None

    subtitle = re.sub("\d+\.+:*", ":", title)
    subtitle = re.sub("\s+-\s+", " : ", subtitle)

    semicolonPosition = subtitle.find(":")
    if -1 != semicolonPosition:
        subtitle = subtitle[semicolonPosition + 1:]

    subtitle = subtitle.strip()

    return subtitle

def Stringify(value: Any) -> str:

    ##
    #
    # Converts any value to a string.
    #
    # @param value The input value.
    #
    # @return Stringified input value.
    #
    ##

    if value is None:
        return ""

    elif isinstance(value, str):
        return value

    elif isinstance(value, bytes):
        return value.decode("utf-8")

    else:
        return str(value)

def Truncate(string: str, length: int, suffix: str = "…") -> str:

    ##
    #
    # Truncates a string to given length. Adds ellipsis at the end (if necessary).
    #
    # @param string The string to be truncated.
    # @param length Maximum length of the output string.
    # @param suffix The suffix of the output string.
    #
    # @return The truncated string.
    #
    ##

    if len(string) <= length:
        return string

    return string[:length - len(suffix)].rstrip() + suffix

def TruncateByWords(string: str, length: int, suffix: str = "…") -> str:

    ##
    #
    # Truncates a string to given length while trying to avoid cutting words in the middle.
    #
    # @param string The string to be truncated.
    # @param length Maximum length of the output string.
    # @param suffix The suffix of the output string.
    #
    # @return The truncated string.
    #
    ##

    if len(string) <= length:
        return string

    return string[:length - len(suffix)].rsplit(" ", 1)[0] + suffix


####
#
# Dreamy Utilities
# Copyright (C) (2020) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Standard packages.

import html
import re
from typing import Optional

# Non-standard packages.

from bs4 import BeautifulSoup

#
#
#
# Functions.
#
#
#

def EscapeHTMLEntities(code: str) -> str:

    ##
    #
    # Escapes HTML entities in code.
    #
    # @param code The input code.
    #
    # @return The input code with HTML entities escaped.
    #
    ##

    if not code:
        return code

    code = UnescapeHTMLEntities(code)

    return html.escape(code)

def ReadElementText(soup: BeautifulSoup, tagName: str) -> Optional[str]:

    ##
    #
    # Returns the content of the first found element of given name.
    #
    # @param soup    The input tag soup.
    # @param tagName The name of the tag.
    #
    # @return The content of the found element, or **None**.
    #
    ##

    if (not soup) or (not tagName):
        return None

    element = soup.select_one(tagName)
    if not element:
        return None

    return element.get_text().strip()

def UnescapeHTMLEntities(code: str) -> str:

    ##
    #
    # Unescapes HTML entities.
    #
    # @param code The input code.
    #
    # @return The processed code.
    #
    ##

    if not code:
        return codde

    for entity in re.findall("&[a-z]+?;", code):
        code = code.replace(entity, entity.lower())

    return html.unescape(code)